; ----------------------------- MOTHER32.ASM ---------------------------
; Mother-of-All random number generator by Agner Fog 1998
; 32-bit mode version for 80x86 and compatible microprocessors
;
; MRandom returns a floating point number between 0 and 1.
; MRandomInit must be called before the first call to MRandom.
;
; C++ prototypes:
; extern"C" void _cdecl MRandomInit (int seed);
; extern"C" double _cdecl MRandom (void);
;
; ----------------------------------------------------------------------

; The MRandom function is optimized for the Pentium microprocessor.

.386

_DATA SEGMENT DWORD USE32 PUBLIC 'DATA'
M0      DD      0               ; history buffer
M1      DD      0
M2      DD      0
M3      DD      0
MC      DD      0
MF3     DD      2111111111      ; factors
MF2     DD      1492
MF1     DD      1776
MF0     DD      5115
F2M32   DD      2F800000H       ; float representation of 2^(-32)
EXTEND  DD      0, 0            ; used for conversion
_DATA ENDS


_TEXT   SEGMENT DWORD USE32 PUBLIC 'CODE'
FLAT GROUP _DATA,_TEXT
ASSUME CS:FLAT,DS:FLAT,ES:FLAT

PUBLIC MRandom,_MRandom,MRandomInit,_MRandomInit

MRandom PROC NEAR
_MRandom LABEL NEAR
        PUSH    EDI
        MOV     EAX, [MF3]
        MUL     [M3]             ; X[n-4]
        MOV     ECX,EAX
        MOV     EAX, [M2]        ; X[n-3]
        MOV     EDI,EDX
        MOV     [M3],EAX
        MUL     [MF2]
        ADD     ECX,EAX
        MOV     EAX, [M1]        ; X[n-2]
        ADC     EDI,EDX
        MOV     [M2],EAX
        MUL     [MF1]
        ADD     ECX,EAX
        MOV     EAX,[M0]         ; X[n-1]
        ADC     EDI,EDX
        MOV     [M1],EAX
        MUL     [MF0]
        ADD     EAX,ECX
        ADC     EDX,EDI
        ADD     EAX,[MC]
        ADC     EDX,0
        MOV     [EXTEND], EAX
        MOV     [M0],EAX
        MOV     [MC],EDX
        FILD    QWORD PTR [EXTEND]
        POP     EDI
        FMUL    [F2M32]
        RET
Mrandom ENDP


MRandomInit PROC NEAR
_MRandomInit LABEL NEAR
        MOV     EAX, [ESP+4]    ; seed
        CMP     EAX, 1
        SBB     EAX, 0          ; make sure seed != 0
        XOR     ECX, ECX
        ; make random numbers and put them into buffer
R80:    MOV     EDX, EAX
        SHL     EAX, 13
        XOR     EDX, EAX
        MOV     EAX, EDX
        SHR     EDX, 17
        XOR     EAX, EDX
        MOV     EDX, EAX
        SHL     EDX, 5
        XOR     EAX, EDX
        MOV     [M0][ECX*4], EAX
        INC     ECX
        CMP     ECX, 5
        JB      R80
        PUSH    EDI
        MOV     EDI, 19
R90:    CALL    MRandom
        FSTP    ST(0)
        DEC     EDI
        JNZ     R90
        POP     EDI
        RET     0                  ; RET 4 if not _cdecl calling
MRandomInit ENDP

_TEXT   ENDS
        END


; Example of calling from C++ program:

; #include "arandom.h"
; int i;
; MRandomInit (time(0));
; for (i=0; i<100; i++) {
;   printf ("\n%14.10f", MRandom());}
