; -------------------------- RANROT16.ASM -------------------------------
; random number generator by Agner Fog 1999
; 16-bit mode version for 80386/87, 80486, Pentium and compatible
; microprocessors including math processor.
; Works with DOS and Window 3.x.
;
; The code uses the RANROT type W algorithm without self-test.
; The resolution is 63 bits for WRandom and 32 bits for WBRandom.
;
; WRandomInit must be called before the first call to any of the random
; functions.
; WRandom returns a floating point number between 0 and 1.
; WBRandom returns an integer between 0 and 0xFFFFFFFF
;
; C++ prototypes:
; extern"C" void _cdecl WRandomInit (long seed);
; extern"C" double _cdecl WRandom (void);
; extern"C" unsigned long _cdecl WBRandom (void);
;
; -----------------------------------------------------------------------

.386
; define parameters
KK      equ     17            ; size of circular buffer
JJ      equ     10            ; second pointer
R1      equ     19            ; rotate count
R2      equ     27            ; rotate count

RAND_DATA SEGMENT PARA USE16 PUBLIC 'DATA'
align   16
randp1  DT      1.5           ; used for conversion to float
p1      DW      0             ; pointer in circular buffer
p2      DW      0             ; pointer in circular buffer
align   8
randbuf DD      (2*KK) dup(?) ; circular buffer
one     DD      1.0           ; constant
RAND_DATA ENDS
DGROUP GROUP RAND_DATA

_TEXT   SEGMENT BYTE USE16 PUBLIC 'CODE'
ASSUME CS:_TEXT,DS:DGROUP

PUBLIC WRandom,_WRandom,WBRandom,_WBRandom,WRandomInit,_WRandomInit

WRandom PROC NEAR               ; FAR if model is medium or large
_WRandom LABEL NEAR
        FLD     [randp1]        ; load random number
        FSUB    [one]           ; normalize
Random3 LABEL NEAR              ; second entry for integer output
_Random3 LABEL NEAR
        MOV     BX, [p1]       ; ring buffer pointers
        PUSH    SI
        MOV     SI, [p2]
        NOP                     ; avoid imperfect pairing due to AGI
        MOV     EDX, [randbuf][BX]     ; RANROT algorithm
        MOV     EAX, [randbuf][BX+4]
        ROL     EDX, R1
        ROL     EAX, R2
        ADD     EDX, [randbuf][SI]
        ADD     EAX, [randbuf][SI+4]
        MOV     [randbuf][BX], EAX
        MOV     [randbuf][BX+4], EDX
        OR      EDX, 80000000H  ; bit 63 must be set
        MOV     DWORD PTR [randp1],EAX  ; low 32 bits of next random number
        SUB     SI, 8                   ; decrement p2
        JNC     SHORT R30
        MOV     SI, (KK-1)*8
R30:    MOV     DWORD PTR [randp1+4],EDX ; bits 32-62 of next random number
        SUB     BX, 8                   ; decrement p1
        JNC     SHORT R40
        MOV     BX, (KK-1)*8
R40:    MOV     [p1], BX
        MOV     [p2], SI
        POP     SI
        RET
WRandom ENDP

WBRandom PROC NEAR
_WBRandom LABEL NEAR
        MOV     EAX, DWORD PTR [randp1]
        PUSH    EAX
        CALL    Random3
        POP     AX
        POP     DX
        RET
WBRandom ENDP

WRandomInit PROC NEAR            ; FAR if model is medium or large
_WRandomInit LABEL NEAR
        MOV     BX, SP
        MOV     EAX, DWORD PTR SS:[BX-2*(TYPE WRandomInit)]  ; seed
        CMP     EAX, 1
        SBB     EAX, 0          ; make sure seed != 0

        MOV     BX, OFFSET DS:randbuf    ; set up for loop
        MOV     CX, KK*2
R120:   MOV     EDX, EAX
        SHL     EAX, 13
        XOR     EDX, EAX
        MOV     EAX, EDX
        SHR     EDX, 17
        XOR     EAX, EDX
        MOV     EDX, EAX
        SHL     EDX, 5
        XOR     EAX, EDX
        MOV     DWORD PTR [BX],EAX
        ADD     BX, 4
        DEC     CX
        JNZ     R120
        FLD1
        FSTP    [randp1]        ; initialize bit 64-79
        MOV     [p1], (KK-JJ-1)*8  ; initialize buffer pointers
        MOV     [p2], (KK-1)*8
        CALL    Random3         ; prepare first random number
        RET     0               ; RET 8 if _stdcall or _pascal calling
WRandomInit ENDP

_TEXT   ENDS
        END

; -----------
;
;// Example of calling from C++ program:
;
;   extern "C" void          _cdecl WRandomInit (long);
;   extern "C" double        _cdecl WRandom (void);
;   extern "C" unsigned long _cdecl WBRandom (void);
;
;   int i;
;   WRandomInit (time(0));
;   for (i=0; i<20; i++) {
;     printf ("\n%14.10f", WRandom());}
